/* ===============================================================================================
   SHMUP (shoot'em up) by Nathan Youngman <contact@nathany.com>

   My entry for GameDev.net's SpriteLib contest.   
   Do what you want with this code. Credit me if you like.

   Apr 26, 2000 [Wednesday evening]
				Find out about gamedev.net's SpriteLib contest. A bit late, but decided 
				 to make a game in the little time left.
				Design: simple top-down shoot'em up [very bad Raptor clone] named SHMUP.

   Apr 27, 2000	[Thursday evening]
				Created title screen [title.bmp] and SpriteLib-based palette file [shmup.act]
				Design: plot out tasks.

   Apr 28, 2000	[Friday evening]
				Coding begins.
				Basic windows application: full screen POPUP window 
				DirectX 3, full screen exclusive, 640x480 256 color, double buffered
				Loads uncompressed BMP [title.bmp] and ACT (palette) and displays 
				 until Escape is pressed.

   Apr 29, 2000	[Saturday]				
				Randomly generated map, tile-based wrap-around scrolling
				Player sprite with movement, color key, clipping, 3 frames
				Alien sprites, action functions
				Collision detection, shields, explosions, player bullets				
				Sidebar, display shields, display scores

   Apr 30, 2000 [Sunday]
				Map data (map.h)
				Level data (spawn points)
				Enemies: Bugz, Bat, Alien, Wasp (w/weapon)
				Game Over

   Known Issues
				No timing (limit fps on fast computers)
				Doesn't restore surfaces if Alt+Tab out
				DrawTiles() only supports one row of tiles in source image
				Not the best error checking
*/

#define	DIRECTDRAW_VERSION 0x0300		// DirectDraw 3.0
#define WIN32_LEAN_AND_MEAN

#include <windows.h>
#include <windowsx.h>
#include <commdlg.h>
#include <stdio.h>
#include <ddraw.h>
#include "resource.h"
#include "map.h"

#define APPCLASSNAME			"MyAppClassName"
#define APPWINDOWNAME			"SHMUP"

#define KEY_DOWN(vk_code)		((GetAsyncKeyState(vk_code) & 0x8000) ? 1 : 0)

#define MAX_ALIENS				50
#define MAX_EXPLOSIONS			30
#define MAX_PLAYER_BULLETS		8

// sprite ids
#define SPRITE_PLAYER			0
#define SPRITE_INSECT			1
#define SPRITE_BUGZ				2
#define SPRITE_BAT				3
#define SPRITE_ALIEN			4
#define SPRITE_WASP				5

#define SPRITE_EXPLODE			6
#define SPRITE_EXPLODEBIG		7
#define SPRITE_PULSE			8
#define SPRITE_BULLET			9

#define SPRITE_GAMEOVER			10

#define SPRITE_IDS				11

// game states
#define GAMESTATE_TITLELOAD		1	// load and display title screen
#define	GAMESTATE_TITLEDISPLAY	2
#define GAMESTATE_TITLEWAIT		3	// wait for key press
#define GAMESTATE_GAMELOAD		4
#define GAMESTATE_GAMESTART		5
#define GAMESTATE_GAMELOOP		6
#define GAMESTATE_GAMEUNLOAD	7

/* ===============================================================================================
	Sprite structures
=============================================================================================== */
struct BITMAPGRID {
	int x, y, width, height;
};

// ACTION function pointer
struct SPRITEOBJ;
typedef void ACTION(SPRITEOBJ &sprite);

struct SPRITE {			// common data for sprite objects of the same type
	int id;
	int shields_start;
	int damage;			// damage inflicted from collision
	int points;			// points rewarded for destruction
	
	RECT boundingrect;	// bounding rectangle for collisions (frame independant)
	LPDIRECTDRAWSURFACE2 lpdds;	
	BITMAPGRID * bitmapgrid;
	ACTION * action_start;
	ACTION * action_destroy;
};

struct SPRITEOBJ {
	int x, y;			// location on screen (not world coordinates)
	int xv, yv;			// x,y velocity [not always used]
	int frame, timer;
	int shields;
	bool active;
	ACTION * action;
	SPRITE * info;
};

/* ===============================================================================================
	Global Variables
=============================================================================================== */
HINSTANCE hAppInstance;
HWND hAppWnd;
char ErrorMsg[256] = { NULL };		// if string is defined, displays MessageBox on exit
bool WindowClosed = false;			// used to ensure Game_Main isn't called after WM_CLOSE message is posted

DWORD Score, HighScore = 0;

LPDIRECTDRAW2 lpDD = NULL;
LPDIRECTDRAWSURFACE2 lpDDSPrimary = NULL;
LPDIRECTDRAWSURFACE2 lpDDSBackBuffer = NULL;

LPDIRECTDRAWSURFACE2 lpDDSTitleScreen = NULL;
LPDIRECTDRAWSURFACE2 lpDDSTiles = NULL;
LPDIRECTDRAWSURFACE2 lpDDSSprites = NULL;

LPDIRECTDRAWPALETTE lpDDPalette = NULL;

SPRITE SpriteInfo[SPRITE_IDS];		// info on each type of sprite

SPRITEOBJ Player;
SPRITEOBJ Aliens[MAX_ALIENS];
SPRITEOBJ Explosions[MAX_EXPLOSIONS];
SPRITEOBJ PlayerBullets[MAX_PLAYER_BULLETS];
SPRITEOBJ GameOver;

int MapYTile,MapYOffset;

int GameState = GAMESTATE_TITLELOAD;

/* ===============================================================================================
	BitmapGrid (offsets into sprites.bmp)
=============================================================================================== */
BITMAPGRID PlayerBitmaps[] = { 
	{0,0,94,70},
	{94,0,94,70},
	{188,0,94,70}
};

BITMAPGRID InsectBitmaps[] = { 
	{0,70,34,26},
	{34,70,34,26},
	{68,70,34,26},
	{102,70,34,26},
	{136,70,34,26}
};

BITMAPGRID BugzBitmaps[] = {
	{170,70,32,20},
	{202,70,32,20}
};

BITMAPGRID BatBitmaps[] = {
	{234,70,60,30}
};

BITMAPGRID AlienBitmaps[] = {
	{170,100,30,32},
	{200,100,30,32},
	{230,100,30,32},
	{260,100,30,32},
	{290,100,30,32},
	{320,100,30,32},
	{350,100,30,32},
	{380,100,30,32}
};

BITMAPGRID WaspBitmaps[] = {
	{  0,174,26,20},
	{ 26,174,26,20},
	{ 52,174,26,20},
	{ 78,174,26,20},
	{104,174,26,20},
	{130,174,26,20},
	{156,174,26,20},
	{182,174,26,20},
	{208,174,26,20},
	{234,174,26,20},
	{260,174,26,20},
	{286,174,26,20}
};

BITMAPGRID ExplodeBitmaps[] = {
	{0,100,32,32},
	{32,100,32,32},
	{64,100,32,32},
	{96,100,32,32},
	{128,100,32,32}
};

BITMAPGRID PulseBitmaps[] = {
	{160,100,10,16},
	{160,116,10,16}
};

BITMAPGRID BulletBitmaps[] = {
	{312,174,10,10}
};

BITMAPGRID ExplodeBigBitmaps[] = {
	{0,132,48,42},
	{48,132,48,42},
	{96,132,48,42},
	{144,132,48,42},
	{192,132,48,42},
	{240,132,48,42}
};

BITMAPGRID SidebarBitmaps[] = {
	{480,0,160,480}
};

BITMAPGRID DigitBitmaps[] = {
	{0,300,12,16},
	{12,300,12,16},
	{24,300,12,16},
	{36,300,12,16},
	{48,300,12,16},
	{60,300,12,16},
	{72,300,12,16},
	{84,300,12,16},
	{96,300,12,16},
	{108,300,12,16}	
};

BITMAPGRID GameOverBitmaps[] = {
	{0,317,370,75}
};

/* ===============================================================================================
	Level data
=============================================================================================== */
struct SPAWNPOINTS {
	int time;					// MapYTile * 32 + MapYOffset (decreases as level goes on)
	int x, y;					// screen coordinates
	int id;						// alien to spawn
};

SPAWNPOINTS SpawnPoints[] = {
	{ 1440,480,260,SPRITE_INSECT},
	{ 1434,480,200,SPRITE_INSECT},
	{ 1428,480,140,SPRITE_INSECT},
	{ 1280, 20,  0,SPRITE_BUGZ},
	{ 1279, 60,  0,SPRITE_BUGZ},
	{ 1278,100,  0,SPRITE_BUGZ},
	{ 1270,420,  0,SPRITE_BUGZ},
	{ 1269,380,  0,SPRITE_BUGZ},
	{ 1268,340,  0,SPRITE_BUGZ},
	{ 1260,180,  0,SPRITE_BUGZ},
	{ 1260,260,  0,SPRITE_BUGZ},
	{ 1220,  0,100,SPRITE_BAT},
	{ 1200,480,100,SPRITE_BAT},
	{ 1120,  0,360,SPRITE_BAT},
	{ 1100,480,360,SPRITE_BAT},
	{ 1003,  0, 50,SPRITE_ALIEN},
	{ 1003,480,400,SPRITE_ALIEN},
	{ 1002,  0,100,SPRITE_ALIEN},
	{ 1002,480,350,SPRITE_ALIEN},
	{ 1001,  0,150,SPRITE_ALIEN},
	{ 1001,480,300,SPRITE_ALIEN},
	{ 1000,  0,200,SPRITE_ALIEN},	
	{ 1000,480,250,SPRITE_ALIEN},
	{  800,  0, 60,SPRITE_WASP},
	{  800,480, 60,SPRITE_WASP},
	{  800, 60,  0,SPRITE_WASP},
	{  800,420,  0,SPRITE_WASP},
	{  580,420,  0,SPRITE_BUGZ},
	{  579,380,  0,SPRITE_BUGZ},
	{  578,340,  0,SPRITE_BUGZ},
	{  570, 20,  0,SPRITE_BUGZ},
	{  569, 60,  0,SPRITE_BUGZ},
	{  568,100,  0,SPRITE_BUGZ},	
	{  560,180,  0,SPRITE_BUGZ},
	{  560,260,  0,SPRITE_BUGZ},
	{  500,  0,160,SPRITE_WASP},
	{  500,480,160,SPRITE_WASP},		
	{  500,190,  0,SPRITE_WASP},
	{  500,270,  0,SPRITE_WASP},
	{  340,480,340,SPRITE_INSECT},
	{  334,480,280,SPRITE_INSECT},
	{  328,480,220,SPRITE_INSECT},
	{  128,  0,  0,SPRITE_ALIEN},	
	{  128,480,  0,SPRITE_ALIEN},
	{   96,  0, 32,SPRITE_ALIEN},	
	{   96,480, 32,SPRITE_ALIEN},
	{   64,  0, 64,SPRITE_ALIEN},	
	{   64,480, 64,SPRITE_ALIEN},	
	{    0,  0, 60,SPRITE_WASP},	
	{    0,  0,160,SPRITE_WASP},
	{    0, 80,  0,SPRITE_WASP},
	{999999999,0,0,0}			// terminate
};

int SpawnIndex;

/* ===============================================================================================
	Function Prototypes
=============================================================================================== */
void Game_Main();
bool DDraw_Init();
void DDraw_Shutdown();
void ClearScreen(int color = 0);
void FlipScreen();

LRESULT CALLBACK WindowProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

LPDIRECTDRAWSURFACE2 LoadBMP(char * filename,int mem_flags = 0);
LPDIRECTDRAWPALETTE LoadACT(char * filename);

void DrawTiles();
void CreateMap();

void InitSpriteInfo();
void BltSprite(SPRITEOBJ &sprite);
void InitSprite(SPRITEOBJ &sprite, int x, int y, int id );
void RemoveSprite(SPRITEOBJ &sprite);
bool DetectCollision(SPRITEOBJ &sprite1, SPRITEOBJ &sprite2);
int GetNewSpriteIndex(SPRITEOBJ sprite[],int max);

void DrawSidebar();
void AdjustScore(int points);
void DisplayNumber(int x, int y, int num, int digits);

void ActionPlayer(SPRITEOBJ &sprite);
void ActionPlayerDestroy(SPRITEOBJ &sprite);

void ActionInsect(SPRITEOBJ &sprite);
void ActionBugz(SPRITEOBJ &sprite);
void ActionBat(SPRITEOBJ &sprite);
void ActionBatFlyAway(SPRITEOBJ &sprite);
void ActionAlien(SPRITEOBJ &sprite);
void ActionWasp(SPRITEOBJ &sprite);
void ActionWaspCircleCW(SPRITEOBJ &sprite);
void ActionWaspCircleCCW(SPRITEOBJ &sprite);

void ActionAlienDestroy(SPRITEOBJ &sprite);
void ActionBigAlienDestroy(SPRITEOBJ &sprite);
void ActionExplode(SPRITEOBJ &sprite);
void ActionExplodeBig(SPRITEOBJ &sprite);
void ActionPulse(SPRITEOBJ &sprite);
void ActionBullet(SPRITEOBJ &sprite);

void ActionGameOver(SPRITEOBJ &sprite);

/* ===============================================================================================
	WinMain
=============================================================================================== */
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{	
	MSG msg;

	hAppInstance = hInstance;

	// register new window class
	WNDCLASSEX wcx;
	wcx.cbSize = sizeof(WNDCLASSEX);
	wcx.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC | CS_DBLCLKS;
	wcx.lpfnWndProc = WindowProc;
	wcx.cbClsExtra = wcx.cbWndExtra = 0;
	wcx.hInstance = hAppInstance;
	wcx.hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_APPICON));
	wcx.hCursor = LoadCursor(NULL, IDI_APPLICATION);
	wcx.hbrBackground = GetStockObject(BLACK_BRUSH);
	wcx.lpszMenuName = NULL;
	wcx.lpszClassName = APPCLASSNAME;
	wcx.hIconSm = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_APPICON));
	if(!RegisterClassEx(&wcx)) return(1);
	
	DWORD dwStyle = WS_POPUP | WS_VISIBLE;	
	DWORD dwExStyle = NULL;

	// create window	
	hAppWnd = CreateWindowEx(dwExStyle, APPCLASSNAME, APPWINDOWNAME, dwStyle,		
		0, 0,
		GetSystemMetrics(SM_CXSCREEN), GetSystemMetrics(SM_CYSCREEN),
		HWND_DESKTOP,
		NULL,
		hAppInstance,
		NULL);
	if(!hAppWnd) return(1);

	if( !DDraw_Init() ) { DDraw_Shutdown(); return(1); }


	// Message loop
	while(TRUE)
	{
		if( PeekMessage(&msg, NULL,0,0,PM_REMOVE) )
		{
			if(msg.message == WM_QUIT) break;
			TranslateMessage(&msg);	// translate accelerator keys
			DispatchMessage(&msg);
		}

		Game_Main();
	}

	DDraw_Shutdown();

	if( strlen(ErrorMsg) ) MessageBox(HWND_DESKTOP,ErrorMsg,"Error",MB_OK | MB_ICONSTOP);
	
	return(msg.wParam);	
}

/* ===============================================================================================
	WindowProc
=============================================================================================== */
LRESULT CALLBACK WindowProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{	
	switch(uMsg) 
	{
		case WM_CREATE:
		{
			// init stuff here			
		}
		break;
		
		case WM_PAINT:
		{			
			ValidateRect(hWnd,NULL);
		}
		break;	
		
		case WM_DESTROY:
		{
			PostQuitMessage(0);			
		}
		break;
	
		default:
		{
			return( DefWindowProc(hWnd, uMsg, wParam, lParam) );
		}
	}

	return(0);
}

/* ===============================================================================================
	DDraw_Init 
=============================================================================================== */
bool DDraw_Init()
{
	HRESULT ddrval;
	LPDIRECTDRAW lpDD1;	

	ShowCursor(false);		// hide mouse cursor

	// get DDraw interface
	ddrval = DirectDrawCreate(NULL, &lpDD1, NULL);
	if(ddrval != DD_OK)	{ strcpy(ErrorMsg, "Unable to initialize DirectDraw."); return false; }
		
	// get DDraw2 interface
	ddrval = lpDD1->QueryInterface(IID_IDirectDraw2,(LPVOID *)&lpDD);
	if(ddrval != DD_OK) return false;

	lpDD1->Release(); lpDD1 = NULL;						// release old interface

	ddrval = lpDD->SetCooperativeLevel(hAppWnd,DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN | DDSCL_NOWINDOWCHANGES);
	if(ddrval != DD_OK) return false;

	ddrval = lpDD->SetDisplayMode(640,480,8,0,0);		// set display mode
	if(ddrval != DD_OK) return false;

	DDSURFACEDESC ddsd;
	LPDIRECTDRAWSURFACE lpDDSPrimary1;
	
	ZeroMemory(&ddsd,sizeof(ddsd));						// creating flipping chain
	ddsd.dwSize = sizeof(ddsd);
	ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
	ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_FLIP | DDSCAPS_COMPLEX;
	ddsd.dwBackBufferCount = 1;
	ddrval = lpDD->CreateSurface(&ddsd, &lpDDSPrimary1, NULL);
	if(ddrval != DD_OK) return false;

	ddrval = lpDDSPrimary1->QueryInterface(IID_IDirectDrawSurface2,(LPVOID *)&lpDDSPrimary);
	if(ddrval != DD_OK) return false;

	lpDDSPrimary1->Release(); lpDDSPrimary1 = NULL;		// release old interface

	DDSCAPS ddscaps;									// get back buffer
	ZeroMemory(&ddscaps,sizeof(ddscaps));
	ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
	ddrval = lpDDSPrimary->GetAttachedSurface(&ddscaps, &lpDDSBackBuffer);

	return true;
}

/* ===============================================================================================
	DDraw_Shutdown
=============================================================================================== */
void DDraw_Shutdown()
{
	if(lpDD != NULL)
    {
		lpDD->SetCooperativeLevel(hAppWnd,DDSCL_NORMAL);

		if(lpDDSPrimary != NULL)
		{
			lpDDSPrimary->Release();
			lpDDSPrimary = NULL;
		}

        lpDD->Release();
        lpDD = NULL;
    }

	ShowCursor(true);
}

/* ===============================================================================================
	LoadBMP
	load uncompressed 256 bitmap to off-screen surface, ignores palette
	mem_flags = DDSCAPS_SYSTEMMEMORY | DDSCAPS_VIDEOMEMORY [default = 0, unspecified, optimal?]
=============================================================================================== */
LPDIRECTDRAWSURFACE2 LoadBMP(char * filename,int mem_flags)
{
	FILE * fp;
	BITMAPFILEHEADER bitmapfileheader;
	BITMAPINFOHEADER bitmapinfoheader;
	HRESULT ddrval;

	fp = fopen(filename, "rb");
	if(!fp) { sprintf(ErrorMsg,"LoadBMP: %s not found",filename); return NULL; }

	fread(&bitmapfileheader,sizeof(bitmapfileheader),1,fp);
	if(bitmapfileheader.bfType != 0x4D42) { fclose(fp); return NULL; }			// not a .BMP

	fread(&bitmapinfoheader,sizeof(bitmapinfoheader),1,fp);
	if(bitmapinfoheader.biBitCount != 8) { fclose(fp); return NULL; }			// not 256 colors
	if(bitmapinfoheader.biCompression != BI_RGB) { fclose(fp); return NULL; }	// not uncompressed
	if(bitmapinfoheader.biHeight < 0) { fclose(fp); return NULL; }				// not upside-down

	// [skip the palette]

	// biSizeImage could = 0, so set it to be sure
	bitmapinfoheader.biSizeImage = bitmapinfoheader.biWidth * bitmapinfoheader.biHeight;
	
	fseek(fp,bitmapfileheader.bfOffBits,SEEK_SET);				// go to bitmap data

	BYTE * buffer;
	buffer = (BYTE *)malloc(bitmapinfoheader.biSizeImage);
	if(!buffer) { fclose(fp); return NULL; }					// out of memeory

	fread(buffer,bitmapinfoheader.biSizeImage,1,fp);			// read bitmap data

	fclose(fp);

	// make off screen surface
	DDSURFACEDESC ddsd;
	LPDIRECTDRAWSURFACE lpdds1;
	LPDIRECTDRAWSURFACE2 lpdds;
	
	ZeroMemory(&ddsd,sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT | DDSD_CKSRCBLT;
	ddsd.dwWidth = bitmapinfoheader.biWidth;
	ddsd.dwHeight = bitmapinfoheader.biHeight;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | mem_flags;

	ddsd.ddckCKSrcBlt.dwColorSpaceLowValue = 0;				// set up color key (index 0)
	ddsd.ddckCKSrcBlt.dwColorSpaceHighValue = 0;

	ddrval = lpDD->CreateSurface(&ddsd,&lpdds1,NULL);		// create surface
	if(ddrval != DD_OK) { free(buffer); strcpy(ErrorMsg,"LoadBMP: unable to create ddrawsurface"); return NULL; }
	
	ddrval = lpdds1->QueryInterface(IID_IDirectDrawSurface2,(LPVOID *)&lpdds);	// get new interface
	lpdds1->Release(); lpdds1 = NULL;						// release old interface
	if(ddrval != DD_OK) { free(buffer); strcpy(ErrorMsg,"LoadBMP: unable to create ddrawsurface2"); return NULL; }


	ZeroMemory(&ddsd,sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	ddrval = lpdds->Lock(NULL, &ddsd, DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL);
	if(ddrval != DD_OK) { lpdds->Release(); lpdds = NULL; free(buffer); strcpy(ErrorMsg,"LoadBMP: unable to lock surface"); return NULL; }
	
	int x,y;
	BYTE * surface = (BYTE *)ddsd.lpSurface;

	for(y = 0; y < bitmapinfoheader.biHeight; y++) 
	{
		for(x = 0; x < bitmapinfoheader.biWidth; x++)
		{
			surface[x+y*ddsd.lPitch] = (BYTE)buffer[x+(bitmapinfoheader.biHeight-y-1)*bitmapinfoheader.biWidth];
		}
	}

	ddrval = lpdds->Unlock(NULL);
	if(ddrval != DD_OK) { lpdds->Release(); lpdds = NULL; free(buffer); strcpy(ErrorMsg,"LoadBMP: unable to unlock surface"); return NULL; }

	free(buffer);

	return lpdds;
}


/* ===============================================================================================
	LoadACT [adobe color table]
	load 256 color palette
=============================================================================================== */
LPDIRECTDRAWPALETTE LoadACT(char * filename)
{
	FILE * fp;
	HRESULT ddrval;
	PALETTEENTRY palette[256];
	LPDIRECTDRAWPALETTE lpddp;

	fp = fopen(filename, "rb");
	if(!fp) { sprintf(ErrorMsg,"LoadACT: %s not found",filename); return NULL; }
	
	for(int z = 0; z < 255; z++)
	{		
		palette[z].peRed = fgetc(fp);
		palette[z].peGreen = fgetc(fp);
		palette[z].peBlue = fgetc(fp);
		palette[z].peFlags = PC_NOCOLLAPSE;
	}
	fclose(fp);

	ddrval = lpDD->CreatePalette(DDPCAPS_8BIT,palette,&lpddp, NULL);
	if(ddrval != DD_OK) return NULL;

	return lpddp;
}

/* ===============================================================================================
	FillScreen
	fills back buffer with solid color
	color = color index (0-255, default is 0)
=============================================================================================== */
void ClearScreen(int color)
{
	DDBLTFX ddbltfx;
	RECT dest_rect;

	ZeroMemory(&ddbltfx,sizeof(ddbltfx));
	ddbltfx.dwSize = sizeof(ddbltfx);

	ddbltfx.dwFillColor = color;
	dest_rect.left = 0; dest_rect.top = 0;
	dest_rect.right = 640; dest_rect.bottom = 480;

	lpDDSBackBuffer->Blt(&dest_rect, NULL, NULL, DDBLT_COLORFILL | DDBLT_WAIT, &ddbltfx);
}

/* ===============================================================================================
	FlipScreen	
=============================================================================================== */
void FlipScreen()
{
	HRESULT ddrval;

	do 
	{
		ddrval = lpDDSPrimary->Flip(NULL,DDFLIP_WAIT);
	}
	while(ddrval != DD_OK);
}

/* ===============================================================================================
	CreateMap
=============================================================================================== */
void CreateMap()
{
	int x, y;

	for(y = 0; y < MAP_HEIGHT; y++) 
	{
		for(x = 0; x < 15; x++) 
		{
			Map[y][x] = rand() % 16;
		}
	}

	MapYTile = MAP_HEIGHT - 15;
	MapYOffset = 0;
}

/* ===============================================================================================
	DrawTiles
	draws 15x15 section of map, implements y-clipping to screen
	negative MapYOffset values allow for wrap-around scrolling	
=============================================================================================== */
void DrawTiles()
{
	int x,y,ypos,tile,map_row,rows;
	RECT src_rect;

	map_row = MapYTile;
	if(map_row < 0) map_row += MAP_HEIGHT;
	
	rows = 15; if(MapYOffset != 0) rows++;	// could actually be 16 rows

	for(y = 0; y < rows; y++) 		
	{
		for(x = 0; x < 15; x++) 
		{
			tile = Map[map_row][x];
			src_rect.top = 0; 
			src_rect.bottom = 32;
			src_rect.left = tile * 32;
			src_rect.right = src_rect.left + 32;
			
			ypos = (y * 32) - MapYOffset;
			if(ypos < 0) 
			{				
				src_rect.top = -ypos; 
				ypos = 0;
			}
			else if(ypos > 448) 
			{
				src_rect.bottom = 480 - ypos;
			}

			lpDDSBackBuffer->BltFast(x*32,ypos,lpDDSTiles,&src_rect,DDBLTFAST_WAIT);
		}

		map_row = (map_row + 1) % MAP_HEIGHT;
	}

}

/* ===============================================================================================
	InitSpriteInfo
	init data in SpriteInfo[] array
=============================================================================================== */
void InitSpriteInfo()
{
	int el;

	ZeroMemory(&SpriteInfo,sizeof(SpriteInfo));

	el = SPRITE_PLAYER;
	SpriteInfo[el].id = el;
	SpriteInfo[el].shields_start = 100;
	SpriteInfo[el].damage = 10;
	SpriteInfo[el].points = 0;					// no points for dying 

	SpriteInfo[el].lpdds = lpDDSSprites;
	SpriteInfo[el].bitmapgrid = PlayerBitmaps;
	
	SpriteInfo[el].boundingrect.left = 30; 
	SpriteInfo[el].boundingrect.right = 64;		// redefine bounding box for player
	SpriteInfo[el].boundingrect.top = 0; 
	SpriteInfo[el].boundingrect.bottom = 60;	

	SpriteInfo[el].action_start = ActionPlayer;
	SpriteInfo[el].action_destroy = ActionPlayerDestroy;

	// insect
	el = SPRITE_INSECT;
	SpriteInfo[el].id = el;
	SpriteInfo[el].shields_start = 10;
	SpriteInfo[el].damage = 10;
	SpriteInfo[el].points = 50;

	SpriteInfo[el].lpdds = lpDDSSprites;
	SpriteInfo[el].bitmapgrid = InsectBitmaps;
	
	SpriteInfo[el].boundingrect.left = 0;
	SpriteInfo[el].boundingrect.right = 34;
	SpriteInfo[el].boundingrect.top = 0; 
	SpriteInfo[el].boundingrect.bottom = 26;	

	SpriteInfo[el].action_start = ActionInsect;
	SpriteInfo[el].action_destroy = ActionAlienDestroy;

	// bugz
	el = SPRITE_BUGZ;
	SpriteInfo[el].id = el;
	SpriteInfo[el].shields_start = 10;
	SpriteInfo[el].damage = 10;
	SpriteInfo[el].points = 100;

	SpriteInfo[el].lpdds = lpDDSSprites;
	SpriteInfo[el].bitmapgrid = BugzBitmaps;
	
	SpriteInfo[el].boundingrect.left = 0;
	SpriteInfo[el].boundingrect.right = 32;
	SpriteInfo[el].boundingrect.top = 0; 
	SpriteInfo[el].boundingrect.bottom = 20;	

	SpriteInfo[el].action_start = ActionBugz;
	SpriteInfo[el].action_destroy = ActionAlienDestroy;

	// bat
	el = SPRITE_BAT;
	SpriteInfo[el].id = el;
	SpriteInfo[el].shields_start = 10;
	SpriteInfo[el].damage = 20;
	SpriteInfo[el].points = 300;

	SpriteInfo[el].lpdds = lpDDSSprites;
	SpriteInfo[el].bitmapgrid = BatBitmaps;
	
	SpriteInfo[el].boundingrect.left = 0;
	SpriteInfo[el].boundingrect.right = 60;
	SpriteInfo[el].boundingrect.top = 0; 
	SpriteInfo[el].boundingrect.bottom = 30;	

	SpriteInfo[el].action_start = ActionBat;
	SpriteInfo[el].action_destroy = ActionBigAlienDestroy;

	// alien
	el = SPRITE_ALIEN;
	SpriteInfo[el].id = el;
	SpriteInfo[el].shields_start = 20;
	SpriteInfo[el].damage = 0;					// no collisions
	SpriteInfo[el].points = 100;
	SpriteInfo[el].lpdds = lpDDSSprites;
	SpriteInfo[el].bitmapgrid = AlienBitmaps;	
	SpriteInfo[el].boundingrect.left = 0;
	SpriteInfo[el].boundingrect.right = 30;
	SpriteInfo[el].boundingrect.top = 0; 
	SpriteInfo[el].boundingrect.bottom = 32;
	SpriteInfo[el].action_start = ActionAlien;
	SpriteInfo[el].action_destroy = ActionAlienDestroy;

	// wasp
	el = SPRITE_WASP;
	SpriteInfo[el].id = el;
	SpriteInfo[el].shields_start = 20;
	SpriteInfo[el].damage = 10;
	SpriteInfo[el].points = 200;
	SpriteInfo[el].lpdds = lpDDSSprites;
	SpriteInfo[el].bitmapgrid = WaspBitmaps;
	SpriteInfo[el].boundingrect.left = 0;
	SpriteInfo[el].boundingrect.right = 26;
	SpriteInfo[el].boundingrect.top = 0;
	SpriteInfo[el].boundingrect.bottom = 20;
	SpriteInfo[el].action_start = ActionWasp;
	SpriteInfo[el].action_destroy = ActionAlienDestroy;

	// explode
	el = SPRITE_EXPLODE;
	SpriteInfo[el].id = el;
	SpriteInfo[el].shields_start = 0;
	SpriteInfo[el].damage = 0;
	SpriteInfo[el].points = 0;

	SpriteInfo[el].lpdds = lpDDSSprites;
	SpriteInfo[el].bitmapgrid = ExplodeBitmaps;
	
	SpriteInfo[el].boundingrect.left = 0;
	SpriteInfo[el].boundingrect.right = 32;
	SpriteInfo[el].boundingrect.top = 0; 
	SpriteInfo[el].boundingrect.bottom = 32;

	SpriteInfo[el].action_start = ActionExplode;
	SpriteInfo[el].action_destroy = NULL;

	// explode (big)
	el = SPRITE_EXPLODEBIG;
	SpriteInfo[el].id = el;
	SpriteInfo[el].shields_start = 0;
	SpriteInfo[el].damage = 0;
	SpriteInfo[el].points = 0;

	SpriteInfo[el].lpdds = lpDDSSprites;
	SpriteInfo[el].bitmapgrid = ExplodeBigBitmaps;
	
	SpriteInfo[el].boundingrect.left = 0;
	SpriteInfo[el].boundingrect.right = 48;
	SpriteInfo[el].boundingrect.top = 0; 
	SpriteInfo[el].boundingrect.bottom = 42;

	SpriteInfo[el].action_start = ActionExplodeBig;
	SpriteInfo[el].action_destroy = NULL;

	// pulse
	el = SPRITE_PULSE;
	SpriteInfo[el].id = el;
	SpriteInfo[el].shields_start = 0;
	SpriteInfo[el].damage = 10;
	SpriteInfo[el].points = 0;

	SpriteInfo[el].lpdds = lpDDSSprites;
	SpriteInfo[el].bitmapgrid = PulseBitmaps;
	
	SpriteInfo[el].boundingrect.left = 0;
	SpriteInfo[el].boundingrect.right = 10;
	SpriteInfo[el].boundingrect.top = 0; 
	SpriteInfo[el].boundingrect.bottom = 16;

	SpriteInfo[el].action_start = ActionPulse;
	SpriteInfo[el].action_destroy = NULL;

	// bullet
	el = SPRITE_BULLET;
	SpriteInfo[el].id = el;
	SpriteInfo[el].shields_start = 0;
	SpriteInfo[el].damage = 10;
	SpriteInfo[el].points = 0;
	SpriteInfo[el].lpdds = lpDDSSprites;
	SpriteInfo[el].bitmapgrid = BulletBitmaps;	
	SpriteInfo[el].boundingrect.left = 0;
	SpriteInfo[el].boundingrect.right = 10;
	SpriteInfo[el].boundingrect.top = 0; 
	SpriteInfo[el].boundingrect.bottom = 10;
	SpriteInfo[el].action_start = ActionBullet;
	SpriteInfo[el].action_destroy = ActionAlienDestroy;

	// game over
	el = SPRITE_GAMEOVER;
	SpriteInfo[el].id = el;
	SpriteInfo[el].shields_start = 0;
	SpriteInfo[el].damage = 0;
	SpriteInfo[el].points = 0;
	SpriteInfo[el].lpdds = lpDDSSprites;
	SpriteInfo[el].bitmapgrid = GameOverBitmaps;	
	SpriteInfo[el].boundingrect.left = 0;
	SpriteInfo[el].boundingrect.right = 370;
	SpriteInfo[el].boundingrect.top = 0; 
	SpriteInfo[el].boundingrect.bottom = 75;
	SpriteInfo[el].action_start = ActionGameOver;
	SpriteInfo[el].action_destroy = NULL;
}


/* ===============================================================================================
	BltSprite
	implements clipping to 480x480 region
=============================================================================================== */
void BltSprite(SPRITEOBJ &sprite)
{
	int x, y, width, height;
	RECT src_rect;

	x = sprite.x;
	y = sprite.y;

	src_rect.left = sprite.info->bitmapgrid[sprite.frame].x;
	src_rect.top = sprite.info->bitmapgrid[sprite.frame].y;

	width = sprite.info->bitmapgrid[sprite.frame].width;	
	height = sprite.info->bitmapgrid[sprite.frame].height;

	// do clipping 
	if(x < 0)
	{
		src_rect.left += -x; width -= -x; x = 0;
	}
	else if(x > 480 - width)
	{
		width = 480 - x;
	}	

	if(y < 0)
	{
		src_rect.top += -y; height -= -y; y = 0;
	}
	else if(y > 480 - height)
	{
		height = 480 - y;
	}

	src_rect.right = src_rect.left + width;
	src_rect.bottom = src_rect.top + height;
	
	lpDDSBackBuffer->BltFast(x,y,sprite.info->lpdds,&src_rect,DDBLTFAST_WAIT | DDBLTFAST_SRCCOLORKEY);
}

/* ===============================================================================================
	InitSprite
=============================================================================================== */
void InitSprite(SPRITEOBJ &sprite, int x, int y, int id )
{
	ZeroMemory(&sprite,sizeof(sprite));

	sprite.x = x;
	sprite.y = y;
	sprite.frame = sprite.timer = 0;
	
	sprite.info = &SpriteInfo[id];

	sprite.action = sprite.info->action_start;
	sprite.shields = sprite.info->shields_start;

	sprite.active = true;	
}

/* ===============================================================================================
	RemoveSprite
=============================================================================================== */
void RemoveSprite(SPRITEOBJ &sprite)
{
	sprite.active = false;
	ZeroMemory(&sprite,sizeof(sprite));
}

/* ===============================================================================================
	DetectCollision
	detects collision between 2 sprites using bounding rectangles
=============================================================================================== */
bool DetectCollision(SPRITEOBJ &sprite1, SPRITEOBJ &sprite2)
{
	int xmin1,xmax1,xmin2,xmax2,ymin1,ymax1,ymin2,ymax2;

	if( !sprite1.active || !sprite2.active ) return false;		// must be alive to collide

	xmin1 = sprite1.x + sprite1.info->boundingrect.left;
	xmax1 = sprite1.x + sprite1.info->boundingrect.right;
	ymin1 = sprite1.y + sprite1.info->boundingrect.top;
	ymax1 = sprite1.y + sprite1.info->boundingrect.bottom;

	xmin2 = sprite2.x + sprite2.info->boundingrect.left;
	xmax2 = sprite2.x + sprite2.info->boundingrect.right;
	ymin2 = sprite2.y + sprite2.info->boundingrect.top;
	ymax2 = sprite2.y + sprite2.info->boundingrect.bottom;
	
	if(xmax2 < xmin1) return false;		// sprite 2 entirely to the left of sprite 1
	if(xmin2 > xmax1) return false;		// sprite 2 entirely to the right of sprite 1
	if(ymax2 < ymin1) return false;		// sprite 2 entirely above sprite 1
	if(ymin2 > ymax1) return false;		// sprite 2 entirely below sprite 1
	
	return true;	// objects overlap
}


/* ===============================================================================================
	GetNewSpriteIndex
	searches array for inactive sprite and returns index into array
=============================================================================================== */
int GetNewSpriteIndex(SPRITEOBJ sprite[],int max)
{
	int z;
	
	for(z = 0; z < max; z++) 
	{
		if( !sprite[z].active ) return z;
	}

	return -1;	// error
}

/* ===============================================================================================
	DrawSidebar
=============================================================================================== */
void DrawSidebar()
{
	RECT src_rect;

	// backdrop
	src_rect.left = SidebarBitmaps[0].x;
	src_rect.top = SidebarBitmaps[0].y;
	src_rect.right = src_rect.left + SidebarBitmaps[0].width;
	src_rect.bottom = src_rect.top + SidebarBitmaps[0].height;
	lpDDSBackBuffer->BltFast(480,0,lpDDSSprites,&src_rect,DDBLTFAST_WAIT);

	// shields
	if(Player.active)
	{
		src_rect.left = PlayerBitmaps[1].x;
		src_rect.top = PlayerBitmaps[1].y;
		src_rect.right = src_rect.left + (PlayerBitmaps[1].width * Player.shields / Player.info->shields_start);
		src_rect.bottom = src_rect.top + PlayerBitmaps[1].height;
		if(src_rect.right > src_rect.left) 
		{
			lpDDSBackBuffer->BltFast(500,380,lpDDSSprites,&src_rect,DDBLTFAST_WAIT | DDBLTFAST_SRCCOLORKEY);
		}
	}

	// score
	DisplayNumber(503,128,Score,10);
	DisplayNumber(503,189,HighScore,10);
}


/* ===============================================================================================
	DisplayNumber
=============================================================================================== */
void DisplayNumber(int x, int y, int num, int digits)
{
	int z,n;
	char buf[20],mask[20];
	RECT src_rect;
	
	sprintf(mask,"%%0%dd", digits);
	sprintf(buf, mask, num);		// "%010d"

	for(z = 0; z < (signed)strlen(buf); z++) 
	{
		n = buf[z] - '0';
				
		src_rect.left = DigitBitmaps[n].x;
		src_rect.top = DigitBitmaps[n].y;
		src_rect.right = src_rect.left + DigitBitmaps[0].width;
		src_rect.bottom = src_rect.top + DigitBitmaps[0].height;
		lpDDSBackBuffer->BltFast(x + z*12,y,lpDDSSprites,&src_rect,DDBLTFAST_WAIT | DDBLTFAST_SRCCOLORKEY);
	}
}

/* ===============================================================================================
	AdjustScore
=============================================================================================== */
void AdjustScore(int points)
{
	Score += points;
	if(Score > HighScore) HighScore = Score;
}

/* ===============================================================================================
	Actions
=============================================================================================== */
void ActionPlayer(SPRITEOBJ &sprite)
{
	if( KEY_DOWN(VK_LEFT))
	{
		sprite.x-=4;
		if( sprite.x < -(sprite.info->bitmapgrid[0].width/2) ) sprite.x = -(sprite.info->bitmapgrid[0].width/2);
	}
	else if( KEY_DOWN(VK_RIGHT) )
	{
		sprite.x+=4;
		if( sprite.x > (480 - sprite.info->bitmapgrid[0].width/2) ) sprite.x = 480 - sprite.info->bitmapgrid[0].width/2;
	}

	sprite.frame = 1;		// medium thrust
	if( KEY_DOWN(VK_UP))
	{
		sprite.y-=4;
		if(sprite.y < 0) sprite.y = 0;
	
		sprite.frame = 2;
	}
	else if( KEY_DOWN(VK_DOWN) )
	{
		sprite.y+=4;
		if(sprite.y > (480 - sprite.info->bitmapgrid[0].height)) sprite.y = 480 - sprite.info->bitmapgrid[0].height;
		
		sprite.frame = 0;
	}

	if( KEY_DOWN(VK_SPACE) )
	{			
		if(sprite.timer == 0) {
			int el1, el2;
			
			el1 = GetNewSpriteIndex(PlayerBullets,MAX_PLAYER_BULLETS);			
			if(el1 != -1) InitSprite(PlayerBullets[el1], sprite.x + 18, sprite.y + 10, SPRITE_PULSE);
			
			el2 = GetNewSpriteIndex(PlayerBullets,MAX_PLAYER_BULLETS);
			if(el2 != -1) InitSprite(PlayerBullets[el2], sprite.x + 66, sprite.y + 10, SPRITE_PULSE);
			else RemoveSprite( PlayerBullets[el1] );	// don't shoot either if can only do 1


		}
		sprite.timer = (sprite.timer + 1) % 20;	// every 20 frames that space bar is held down
	}

	
}

void ActionPlayerDestroy(SPRITEOBJ &sprite)
{
	int x,y,el;
	x = sprite.x + (sprite.info->bitmapgrid[sprite.frame].width - ExplodeBigBitmaps[0].width) / 2;
	y = sprite.y + (sprite.info->bitmapgrid[sprite.frame].height - ExplodeBigBitmaps[0].height) / 2;

	el = GetNewSpriteIndex(Explosions,MAX_EXPLOSIONS);
	InitSprite(Explosions[el], x, y, SPRITE_EXPLODEBIG);
	RemoveSprite( sprite );
}

void ActionInsect(SPRITEOBJ &sprite)
{
	sprite.x -= 4;	
	if(sprite.x < -sprite.info->bitmapgrid[sprite.frame].width) RemoveSprite(sprite);	// remove insect
	
	sprite.timer = (sprite.timer + 1) % 4;	// every 4 movements
	if(sprite.timer == 0) sprite.frame = (sprite.frame + 1) % 4;	// rotate frame (4)
}

void ActionBugz(SPRITEOBJ &sprite)
{
	sprite.y += 6;	
	if(sprite.y >= 480 ) RemoveSprite(sprite);	// remove bug
	
	sprite.timer = (sprite.timer + 1) % 4;	// every 4 movements
	if(sprite.timer == 0) sprite.frame = (sprite.frame + 1) % 2;	// rotate frame (2)
}

void ActionBat(SPRITEOBJ &sprite)
{
	if(sprite.x > 212)
	{
		sprite.x -= 4;
		sprite.y += 2;
	}
	else if(sprite.x < 208)
	{
		sprite.x += 4;
		sprite.y += 2;
	}
	else
	{
		sprite.action = ActionBatFlyAway;
	}	
}

void ActionBatFlyAway(SPRITEOBJ &sprite)
{
	sprite.y -= 8;
	if(sprite.y < -sprite.info->bitmapgrid[sprite.frame].height) RemoveSprite(sprite);	// remove bat
}

void ActionAlien(SPRITEOBJ &sprite)
{
	if(sprite.timer == 0) 
	{
		if(sprite.x <= 0)
		{
			sprite.xv = 2; sprite.yv = 0; sprite.frame = 1;
		}
		else if(sprite.x >= 480)
		{
			sprite.xv = -2; sprite.yv = 0; sprite.frame = 2;
		}
		else if(sprite.y <= 0)
		{
			sprite.xv = 0; sprite.yv = 2; sprite.frame = 0;
		}
		else if(sprite.y >= 480)
		{
			sprite.xv = 0; sprite.yv = -2; sprite.frame = 3;
		}
		else
		{
			int d;
			d = rand() % 4;
			switch(d) 
			{
			case 0: sprite.xv = 0; sprite.yv = 2; sprite.frame = 0; break;			
			case 1: sprite.xv = 2; sprite.yv = 0; sprite.frame = 1; break;
			case 2: sprite.xv = -2; sprite.yv = 0; sprite.frame = 2; break;
			case 3: sprite.xv = 0; sprite.yv = -2; sprite.frame = 3; break;
			}
		}
	}
	
	sprite.x += sprite.xv; sprite.y += sprite.yv;
	sprite.y++;	// move down one to stay with map [fake world coordinates]
	if(sprite.x < -10 || sprite.y < -10 || sprite.y > 490 || sprite.x > 490) RemoveSprite(sprite);	// remove alien
	
	sprite.timer = (sprite.timer + 1) % 64;

	if(sprite.timer % 4 == 0)				// rotate walk frames (2)
	{
		if(sprite.frame < 4) sprite.frame += 4;	else sprite.frame -= 4;
	}
}

void ActionWasp(SPRITEOBJ &sprite)
{
	if(sprite.timer == 0)
	{
		// set up starting movements
		if( sprite.x == 0 || (sprite.y == 0 && sprite.x < 227) )
		{
			sprite.xv = 4; sprite.yv = 4; sprite.frame = 4;
		}
		else if(sprite.x == 480 || (sprite.y == 0 && sprite.x >= 227) )
		{
			sprite.xv = -4; sprite.yv = 4; sprite.frame = 6;
		}
	}

	sprite.x += sprite.xv; sprite.y += sprite.yv;
	if(sprite.y > 480 || sprite.x < -30 || sprite.x > 480) RemoveSprite(sprite);

	sprite.timer++;
	if(sprite.timer > 60) 
	{	
		int el;
		el = GetNewSpriteIndex(Aliens,MAX_ALIENS);

		if(sprite.xv == -4) 
		{
			InitSprite(Aliens[el],sprite.x + 6 - 10,sprite.y + 18,SPRITE_BULLET);
			sprite.action = ActionWaspCircleCW;			
		}
		else 
		{
			InitSprite(Aliens[el],sprite.x + 18,sprite.y + 18,SPRITE_BULLET);
			sprite.action = ActionWaspCircleCCW;						
		}
		sprite.timer = 0;		
	}

	if(sprite.timer % 4 == 0)				// rotate wing frames (2)
	{
		if(sprite.frame % 2 == 0) sprite.frame++; else sprite.frame--;
	}
}

void ActionWaspCircleCW(SPRITEOBJ &sprite)
{
	if(sprite.timer == 0)
	{
		if(sprite.xv == -4 && sprite.yv == 4) 
		{
			sprite.yv = -4; sprite.frame = 8;
		}
		else if(sprite.xv == -4 && sprite.yv == -4) 
		{
			sprite.xv = 0; sprite.frame = 0;
		}
		else if(sprite.xv == 0 && sprite.yv == -4)
		{
			sprite.xv = 4; sprite.frame = 10;
		}
		else if(sprite.xv == 4 && sprite.yv == -4)
		{
			sprite.yv = 4; sprite.frame = 4;
		}
		else if(sprite.xv == 4 && sprite.yv == 4)
		{
			sprite.xv = 0; sprite.frame = 2;
		}
		else if(sprite.xv == 0 && sprite.yv == 4)
		{
			sprite.xv = -4; sprite.frame = 6;
			sprite.timer = 0;			
			sprite.action = ActionWasp;	
		}
	}

	sprite.x += sprite.xv; sprite.y += sprite.yv;	
	if(sprite.y > 480 || sprite.x < -30 || sprite.x > 480) RemoveSprite(sprite);

	sprite.timer = (sprite.timer + 1) % 16;	

	if(sprite.timer % 4 == 0)				// rotate wing frames (2)
	{
		if(sprite.frame % 2 == 0) sprite.frame++; else sprite.frame--;
	}
}

void ActionWaspCircleCCW(SPRITEOBJ &sprite)
{
	if(sprite.timer == 0)
	{
		if(sprite.xv == 4 && sprite.yv == 4) 
		{
			sprite.yv = -4; sprite.frame = 10;
		}
		else if(sprite.xv == 4 && sprite.yv == -4) 
		{
			sprite.xv = 0; sprite.frame = 0;
		}
		else if(sprite.xv == 0 && sprite.yv == -4)
		{
			sprite.xv = -4; sprite.frame = 8;
		}
		else if(sprite.xv == -4 && sprite.yv == -4)
		{
			sprite.yv = 4; sprite.frame = 6;
		}
		else if(sprite.xv == -4 && sprite.yv == 4)
		{
			sprite.xv = 0; sprite.frame = 2;
		}
		else if(sprite.xv == 0 && sprite.yv == 4)
		{
			sprite.xv = 4; sprite.frame = 4;
			sprite.timer = 0;			
			sprite.action = ActionWasp;	
		}
	}

	sprite.x += sprite.xv; sprite.y += sprite.yv;
	if(sprite.y > 480 || sprite.x < -30 || sprite.x > 480) RemoveSprite(sprite);

	sprite.timer = (sprite.timer + 1) % 16;
	
	if(sprite.timer % 4 == 0)				// rotate wing frames (2)
	{
		if(sprite.frame % 2 == 0) sprite.frame++; else sprite.frame--;
	}
}

void ActionAlienDestroy(SPRITEOBJ &sprite)
{
	AdjustScore( sprite.info->points );

	int x,y,el;
	x = sprite.x + (sprite.info->bitmapgrid[sprite.frame].width - ExplodeBitmaps[0].width) / 2;
	y = sprite.y + (sprite.info->bitmapgrid[sprite.frame].height - ExplodeBitmaps[0].height) / 2;

	el = GetNewSpriteIndex(Explosions,MAX_EXPLOSIONS);
	InitSprite(Explosions[el], x, y, SPRITE_EXPLODE);
	RemoveSprite( sprite );	
}

void ActionBigAlienDestroy(SPRITEOBJ &sprite)
{
	AdjustScore( sprite.info->points );

	int x,y,el;
	x = sprite.x + (sprite.info->bitmapgrid[sprite.frame].width - ExplodeBigBitmaps[0].width) / 2;
	y = sprite.y + (sprite.info->bitmapgrid[sprite.frame].height - ExplodeBigBitmaps[0].height) / 2;

	el = GetNewSpriteIndex(Explosions,MAX_EXPLOSIONS);
	InitSprite(Explosions[el], x, y, SPRITE_EXPLODEBIG);
	RemoveSprite( sprite );	
}

void ActionExplode(SPRITEOBJ &sprite)
{	
	sprite.timer = (sprite.timer + 1) % 4;			// every 4 frames
	if(sprite.timer == 0) {
		sprite.frame++;								// next animation frame
		if(sprite.frame >= 5) RemoveSprite(sprite);	// remove after last frame
	}
}

void ActionExplodeBig(SPRITEOBJ &sprite)
{	
	sprite.timer = (sprite.timer + 1) % 4;			// every 4 frames
	if(sprite.timer == 0) {
		sprite.frame++;								// next animation frame
		if(sprite.frame >= 6) RemoveSprite(sprite);	// remove after last frame
	}
}

void ActionPulse(SPRITEOBJ &sprite)
{
	sprite.y -= 6;	
	if(sprite.y < -sprite.info->bitmapgrid[sprite.frame].height) RemoveSprite(sprite);
}

void ActionBullet(SPRITEOBJ &sprite)
{
	sprite.y += 6;	
	if(sprite.y > 480) RemoveSprite(sprite);
}

void ActionGameOver(SPRITEOBJ &sprite)
{
	if( KEY_DOWN(VK_SPACE) )
	{
		RemoveSprite(sprite);

		GameState = GAMESTATE_GAMESTART;
	}
}

/* ===============================================================================================
	Game_Main
=============================================================================================== */
void Game_Main()
{
	if(WindowClosed) return;		// abort if main window has been closed

	switch(GameState) 
	{
		case GAMESTATE_TITLELOAD:
		{
			lpDDPalette = LoadACT("shmup.act");			// load palette
			if( !lpDDPalette ) { PostMessage(hAppWnd, WM_CLOSE, 0, 0); WindowClosed = true; return; }
			lpDDSPrimary->SetPalette(lpDDPalette);		// attach to primary surface


			lpDDSTitleScreen = LoadBMP("title.bmp",DDSCAPS_SYSTEMMEMORY);			
			if( !lpDDSTitleScreen ) { PostMessage(hAppWnd, WM_CLOSE, 0, 0); WindowClosed = true; return; }
		
			GameState = GAMESTATE_TITLEDISPLAY;
			break;
		}

		case GAMESTATE_TITLEDISPLAY:
		{
			lpDDSBackBuffer->BltFast(0,0,lpDDSTitleScreen,NULL,DDBLTFAST_WAIT);
			FlipScreen();

			GameState = GAMESTATE_TITLEWAIT;
			break;
		}

		case GAMESTATE_TITLEWAIT:
		{
			if( KEY_DOWN(VK_ESCAPE) )
			{
				ClearScreen();
				FlipScreen();

				lpDDSTitleScreen->Release(); 
				lpDDSTitleScreen = NULL;
				lpDDPalette->Release();
				lpDDPalette = NULL;			
				PostMessage(hAppWnd, WM_CLOSE, 0, 0); WindowClosed = true; return;
			}

			if( KEY_DOWN(VK_SPACE) )
			{
				ClearScreen();
				FlipScreen();

				GameState = GAMESTATE_GAMELOAD;
			}
			break;
		}

		case GAMESTATE_GAMELOAD:
		{
			lpDDSTiles = LoadBMP("tiles.bmp",DDSCAPS_VIDEOMEMORY);
			if( !lpDDSTiles ) { PostMessage(hAppWnd, WM_CLOSE, 0, 0); WindowClosed = true; return; }
			
			lpDDSSprites = LoadBMP("sprites.bmp");
			if( !lpDDSSprites ) { PostMessage(hAppWnd, WM_CLOSE, 0, 0); WindowClosed = true; return; }

			GameState = GAMESTATE_GAMESTART;
		}

		case GAMESTATE_GAMESTART:
		{
			MapYTile = MAP_HEIGHT - 15; MapYOffset = 0;
			SpawnIndex = 0;

			InitSpriteInfo();
			ZeroMemory(Aliens,sizeof(Aliens));
			ZeroMemory(Explosions,sizeof(Explosions));
			ZeroMemory(&GameOver,sizeof(GameOver));

			InitSprite(Player,193,380,SPRITE_PLAYER);

			Score = 0;

			while( KEY_DOWN(VK_SPACE) );	// wait until space bar is released
						
			GameState = GAMESTATE_GAMELOOP;
			break;
		}

		case GAMESTATE_GAMELOOP:
		{
			int z, i, el;
			
			if( KEY_DOWN(VK_ESCAPE) )
			{				
				GameState = GAMESTATE_GAMEUNLOAD;
			}

			// game over
			if(GameOver.active) GameOver.action(GameOver);

			if( !Player.active && !GameOver.active && !KEY_DOWN(VK_SPACE) )
			{
				InitSprite(GameOver,55,200,SPRITE_GAMEOVER);
			}


			// move player, player bullets					
			if(Player.active) Player.action(Player);

			for(z = 0; z < MAX_PLAYER_BULLETS; z++) 
			{
				if( PlayerBullets[z].active ) PlayerBullets[z].action(PlayerBullets[z]);
			}

			// animate explosions
			for(z = 0; z < MAX_EXPLOSIONS; z++) 
			{
				if( Explosions[z].active ) Explosions[z].action(Explosions[z]);
			}

			// spawn enemy sprites
			while( SpawnPoints[SpawnIndex].time == MapYTile * 32 + MapYOffset )
			{
				el = GetNewSpriteIndex(Aliens,MAX_ALIENS);
				InitSprite(Aliens[el],SpawnPoints[SpawnIndex].x,SpawnPoints[SpawnIndex].y,SpawnPoints[SpawnIndex].id);
				SpawnIndex++;
			}

			

			// move aliens, do collision checking
			for(z = 0; z < MAX_ALIENS; z++) 
			{
				if( Aliens[z].active ) Aliens[z].action(Aliens[z]);

				if( Aliens[z].active )	// note: sprite could be removed during action above
				{	
					if( Aliens[z].info->id != SPRITE_BULLET )	// can't shoot enemy bullets
					{
						for(i = 0; i < MAX_PLAYER_BULLETS; i++)
						{
							if( PlayerBullets[i].active && DetectCollision(PlayerBullets[i],Aliens[z]) )
							{
								PlayerBullets[i].shields -= Aliens[z].info->damage;
								Aliens[z].shields -= PlayerBullets[i].info->damage;

								if(PlayerBullets[i].shields <= 0) RemoveSprite(PlayerBullets[i]);
								if(Aliens[z].shields <=0) Aliens[z].info->action_destroy(Aliens[z]);
							}
						}
					}
					
					if( DetectCollision(Player,Aliens[z]) ) 
					{
						if( Aliens[z].info->id != SPRITE_ALIEN )	// can't bang into land-crawlers
						{
							Player.shields -= Aliens[z].info->damage;
							Aliens[z].shields -= Player.info->damage;

							if(Player.shields <= 0) Player.info->action_destroy(Player);
							if(Aliens[z].shields <=0) Aliens[z].info->action_destroy(Aliens[z]);						
						}
					}
				}
			}

			
			

			// draw screen 
			
			DrawTiles();			

			for(z = 0; z < MAX_PLAYER_BULLETS; z++) 
			{
				if( PlayerBullets[z].active ) BltSprite(PlayerBullets[z]);
			}			

			for(z = 0; z < MAX_ALIENS; z++) 
			{
				if( Aliens[z].active ) BltSprite(Aliens[z]);				
			}

			if(Player.active) BltSprite(Player);
	
			for(z = 0; z < MAX_EXPLOSIONS; z++) 
			{
				if( Explosions[z].active ) BltSprite(Explosions[z]);				
			}

			if(GameOver.active) BltSprite(GameOver);

			DrawSidebar();

			FlipScreen();

			// scroll map

			MapYOffset--;
			if(MapYOffset < 0) 
			{ 
				MapYOffset += 32;
				MapYTile--;
				if(MapYTile <= -15) 
				{ 
					MapYTile += MAP_HEIGHT;		// wrap-around
					SpawnIndex = 0;
				}
			}
			break;
		}
		
		case GAMESTATE_GAMEUNLOAD:
		{
			lpDDSSprites->Release(); lpDDSSprites = NULL;
			lpDDSTiles->Release(); lpDDSTiles = NULL;
			
			while( KEY_DOWN(VK_ESCAPE) );	// wait until escape is released

			GameState = GAMESTATE_TITLEDISPLAY;
			break;
		}
	}
}
